@@ -6,12 +6,14 @@ Reference: Weiping Song et al. "AutoInt: Automatic Feature Interaction Learning
 author: massquantity
 
 """
+import os
 from itertools import islice
 import numpy as np
-import tensorflow as tf2
+import pandas as pd
+import tensorflow.compat.v1 as tf
 from tensorflow.keras.initializers import truncated_normal
 from .base import Base, TfMixin
-from ..evaluate.evaluate import EvalMixin
+from ..evaluation.evaluate import EvalMixin
 from ..utils.tf_ops import (
     reg_config,
     dropout_config,
@@ -21,13 +23,19 @@ from ..data.data_generator import DataGenFeat
 from ..utils.sampling import NegativeSampling
 from ..feature import (
     get_predict_indices_and_values,
-    get_recommend_indices_and_values
+    get_recommend_indices_and_values,
+    features_from_dict,
+    add_item_features
 )
-tf = tf2.compat.v1
 tf.disable_v2_behavior()
 
 
 class AutoInt(Base, TfMixin, EvalMixin):
+    user_variables = ["user_feat"]
+    item_variables = ["item_feat"]
+    sparse_variables = ["sparse_feat"]
+    dense_variables = ["dense_feat"]
+
     def __init__(
             self,
             task,
@@ -52,7 +60,7 @@ class AutoInt(Base, TfMixin, EvalMixin):
     ):
         Base.__init__(self, task, data_info, lower_upper_bound)
         TfMixin.__init__(self, tf_sess_config)
-        EvalMixin.__init__(self, task)
+        EvalMixin.__init__(self, task, data_info)
 
         self.task = task
         self.data_info = data_info
@@ -84,8 +92,10 @@ class AutoInt(Base, TfMixin, EvalMixin):
             self.sparse_field_size = self._sparse_field_size(data_info)
         if self.dense:
             self.dense_field_size = self._dense_field_size(data_info)
+        self.all_args = locals()
 
     def _build_model(self):
+        self.graph_built = True
         tf.set_random_seed(self.seed)
         self.labels = tf.placeholder(tf.float32, shape=[None])
         self.is_training = tf.placeholder_with_default(False, shape=[])
@@ -110,13 +120,13 @@ class AutoInt(Base, TfMixin, EvalMixin):
 
         user_feat = tf.get_variable(
             name="user_feat",
-            shape=[self.n_users, self.embed_size],
+            shape=[self.n_users + 1, self.embed_size],
             initializer=truncated_normal(0.0, 0.01),
             regularizer=self.reg
         )
         item_feat = tf.get_variable(
             name="item_feat",
-            shape=[self.n_items, self.embed_size],
+            shape=[self.n_items + 1, self.embed_size],
             initializer=truncated_normal(0.0, 0.01),
             regularizer=self.reg
         )
@@ -215,7 +225,7 @@ class AutoInt(Base, TfMixin, EvalMixin):
         outputs = tf.nn.relu(outputs)
         return outputs
 
-    def _build_train_ops(self, global_steps=None):
+    def _build_train_ops(self, **kwargs):
         if self.task == "rating":
             self.loss = tf.losses.mean_squared_error(labels=self.labels,
                                                      predictions=self.output)
@@ -231,6 +241,13 @@ class AutoInt(Base, TfMixin, EvalMixin):
         else:
             total_loss = self.loss
 
+        if self.lr_decay:
+            n_batches = int(self.data_info.data_size / self.batch_size)
+            self.lr, global_steps = lr_decay_config(self.lr, n_batches,
+                                                    **kwargs)
+        else:
+            global_steps = None
+
         optimizer = tf.train.AdamOptimizer(self.lr)
         optimizer_op = optimizer.minimize(total_loss, global_step=global_steps)
         update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
@@ -240,15 +257,9 @@ class AutoInt(Base, TfMixin, EvalMixin):
     def fit(self, train_data, verbose=1, shuffle=True,
             eval_data=None, metrics=None, **kwargs):
         self.show_start_time()
-        if self.lr_decay:
-            n_batches = int(len(train_data) / self.batch_size)
-            self.lr, global_steps = lr_decay_config(self.lr, n_batches,
-                                                    **kwargs)
-        else:
-            global_steps = None
-
-        self._build_model()
-        self._build_train_ops(global_steps)
+        if not self.graph_built:
+            self._build_model()
+            self._build_train_ops(**kwargs)
 
         if self.task == "ranking" and self.batch_sampling:
             self._check_has_sampled(train_data, verbose)
@@ -264,22 +275,30 @@ class AutoInt(Base, TfMixin, EvalMixin):
                                          self.sparse,
                                          self.dense)
 
-        self.train_feat(data_generator, verbose, shuffle, eval_data, metrics)
+        self.train_feat(data_generator, verbose, shuffle, eval_data, metrics,
+                        **kwargs)
+        self.assign_oov()
 
-    def predict(self, user, item):
-        user = np.asarray(
-            [user]) if isinstance(user, int) else np.asarray(user)
-        item = np.asarray(
-            [item]) if isinstance(item, int) else np.asarray(item)
+    def predict(self, user, item, feats=None, cold_start="average",
+                inner_id=False):
+        user, item = self.convert_id(user, item, inner_id)
+        unknown_num, unknown_index, user, item = self._check_unknown(user, item)
 
-        unknown_num, unknown_index, user, item = self._check_unknown(
-            user, item)
-
-        (user_indices,
-         item_indices,
-         sparse_indices,
-         dense_values) = get_predict_indices_and_values(
+        (
+            user_indices,
+            item_indices,
+            sparse_indices,
+            dense_values
+        ) = get_predict_indices_and_values(
             self.data_info, user, item, self.n_items, self.sparse, self.dense)
+
+        if feats is not None:
+            assert isinstance(feats, (dict, pd.Series)), (
+                "feats must be dict or pandas.Series.")
+            assert len(user_indices) == 1, "only support single user for feats"
+            sparse_indices, dense_values = features_from_dict(
+                self.data_info, sparse_indices, dense_values, feats, "predict")
+
         feed_dict = self._get_feed_dict(user_indices, item_indices,
                                         sparse_indices, dense_values,
                                         None, False)
@@ -290,21 +309,43 @@ class AutoInt(Base, TfMixin, EvalMixin):
         elif self.task == "ranking":
             preds = 1 / (1 + np.exp(-preds))
 
-        if unknown_num > 0:
+        if unknown_num > 0 and cold_start == "popular":
             preds[unknown_index] = self.default_prediction
-
         return preds
 
-    def recommend_user(self, user, n_rec, **kwargs):
-        user = self._check_unknown_user(user)
-        if not user:
-            return   # popular ?
+    def recommend_user(self, user, n_rec, user_feats=None, item_data=None,
+                       cold_start="average", inner_id=False):
+        user_id = self._check_unknown_user(user, inner_id)
+        if user_id is None:
+            if cold_start == "average":
+                user_id = self.n_users
+            elif cold_start == "popular":
+                return self.data_info.popular_items[:n_rec]
+            else:
+                raise ValueError(user)
+
+        (
+            user_indices,
+            item_indices,
+            sparse_indices,
+            dense_values
+        ) = get_recommend_indices_and_values(
+            self.data_info, user_id, self.n_items, self.sparse, self.dense)
+
+        if user_feats is not None:
+            assert isinstance(user_feats, (dict, pd.Series)), (
+                "feats must be dict or pandas.Series.")
+            sparse_indices, dense_values = features_from_dict(
+                self.data_info, sparse_indices, dense_values, user_feats,
+                "recommend")
+        if item_data is not None:
+            assert isinstance(item_data, pd.DataFrame), (
+                "item_data must be pandas DataFrame")
+            assert "item" in item_data.columns, (
+                "item_data must contain 'item' column")
+            sparse_indices, dense_values = add_item_features(
+                self.data_info, sparse_indices, dense_values, item_data)
 
-        (user_indices,
-         item_indices,
-         sparse_indices,
-         dense_values) = get_recommend_indices_and_values(
-            self.data_info, user, self.n_items, self.sparse, self.dense)
         feed_dict = self._get_feed_dict(user_indices, item_indices,
                                         sparse_indices, dense_values,
                                         None, False)
@@ -312,16 +353,16 @@ class AutoInt(Base, TfMixin, EvalMixin):
         recos = self.sess.run(self.output, feed_dict)
         if self.task == "ranking":
             recos = 1 / (1 + np.exp(-recos))
-
-        consumed = self.user_consumed[user]
+        consumed = set(self.user_consumed[user_id])
         count = n_rec + len(consumed)
         ids = np.argpartition(recos, -count)[-count:]
         rank = sorted(zip(ids, recos[ids]), key=lambda x: -x[1])
-        return list(
-            islice(
-                (rec for rec in rank if rec[0] not in consumed), n_rec
-            )
+        recs_and_scores = islice(
+            (rec if inner_id else (self.data_info.id2item[rec[0]], rec[1])
+             for rec in rank if rec[0] not in consumed),
+            n_rec
         )
+        return list(recs_and_scores)
 
     @staticmethod
     def _att_config(att_embed_size):
@@ -335,4 +376,19 @@ class AutoInt(Base, TfMixin, EvalMixin):
             att_layer_num = len(att_embed_size)
         return att_embed_size, att_layer_num
 
+    def save(self, path, model_name, manual=True, inference_only=False):
+        if not os.path.isdir(path):
+            print(f"file folder {path} doesn't exists, creating a new one...")
+            os.makedirs(path)
+        self.save_params(path)
+        if manual:
+            self.save_variables(path, model_name, inference_only)
+        else:
+            self.save_tf_model(path, model_name)
 
+    @classmethod
+    def load(cls, path, model_name, data_info, manual=True):
+        if manual:
+            return cls.load_variables(path, model_name, data_info)
+        else:
+            return cls.load_tf_model(path, model_name, data_info)
