@@ -6,22 +6,20 @@ References: Steffen Rendle et al. "BPR: Bayesian Personalized Ranking from Impli
 author: massquantity
 
 """
-from tensorflow.python.keras import backend as K
-from tensorflow.keras import backend as K
-import time
+import os
 import logging
 from itertools import islice
 from functools import partial
 import numpy as np
-import tensorflow as tf2
+import tensorflow.compat.v1 as tf
 from tensorflow.keras.initializers import (
     zeros as tf_zeros,
     truncated_normal as tf_truncated_normal
 )
 from .base import Base, TfMixin
-from ..evaluate.evaluate import EvalMixin
+from ..evaluation.evaluate import EvalMixin
 from ..utils.sampling import PairwiseSampling
-from ..utils.misc import time_block, colorize
+from ..utils.misc import time_block, assign_oov_vector
 from ..utils.initializers import truncated_normal
 try:
     from ._bpr import bpr_update
@@ -30,7 +28,6 @@ except (ImportError, ModuleNotFoundError):
     logging.basicConfig(format=LOG_FORMAT)
     logging.warning("BPR cython version is not available")
     pass  # may use tf version, then raise error will fail
-tf = tf2.compat.v1
 tf.disable_v2_behavior()
 
 
@@ -38,6 +35,11 @@ class BPR(Base, TfMixin, EvalMixin):
     """
     BPR is only suitable for ranking task
     """
+    user_variables = ["user_embed_var"]
+    item_variables = ["item_embed_var", "item_bias_var"]
+    user_variables_np = ["user_embed"]
+    item_variables_np = ["item_embed"]
+
     def __init__(
             self,
             task="ranking",
@@ -53,7 +55,9 @@ class BPR(Base, TfMixin, EvalMixin):
     ):
 
         Base.__init__(self, task, data_info)
-        EvalMixin.__init__(self, task)
+        EvalMixin.__init__(self, task, data_info)
+        if use_tf:
+            TfMixin.__init__(self)
 
         self.task = task
         self.data_info = data_info
@@ -70,26 +74,24 @@ class BPR(Base, TfMixin, EvalMixin):
         self.user_consumed = data_info.user_consumed
         self.user_embed = None
         self.item_embed = None
-
-        if use_tf:
-            TfMixin.__init__(self)
-        #    self.sess = tf.Session()
-            self._build_model_tf()
-            self._build_train_ops()
-        else:
-            self._build_model()
+        self.all_args = locals()
 
     def _build_model(self):
-        np.random.seed(self.seed)
-        # last dimension is item bias, so for user all set to 1.0
-        self.user_embed = truncated_normal(
-            shape=(self.n_users, self.embed_size + 1), mean=0.0, scale=0.03)
-        self.user_embed[:, self.embed_size] = 1.0
-        self.item_embed = truncated_normal(
-            shape=(self.n_items, self.embed_size + 1), mean=0.0, scale=0.03)
-        self.item_embed[:, self.embed_size] = 0.0
+        if not self.use_tf:
+            self.model_built = True
+            np.random.seed(self.seed)
+            # last dimension is item bias, so for user all set to 1.0
+            self.user_embed = truncated_normal(
+                shape=(self.n_users, self.embed_size + 1), mean=0.0, scale=0.03)
+            self.user_embed[:, self.embed_size] = 1.0
+            self.item_embed = truncated_normal(
+                shape=(self.n_items, self.embed_size + 1), mean=0.0, scale=0.03)
+            self.item_embed[:, self.embed_size] = 0.0
+        else:
+            self._build_model_tf()
 
     def _build_model_tf(self):
+        self.graph_built = True
         if isinstance(self.reg, float) and self.reg > 0.0:
             tf_reg = tf.keras.regularizers.l2(self.reg)
         else:
@@ -101,19 +103,19 @@ class BPR(Base, TfMixin, EvalMixin):
 
         self.item_bias_var = tf.get_variable(
             name="item_bias_var",
-            shape=[self.n_items],
+            shape=[self.n_items + 1],
             initializer=tf_zeros,
             regularizer=tf_reg
         )
         self.user_embed_var = tf.get_variable(
             name="user_embed_var",
-            shape=[self.n_users, self.embed_size],
+            shape=[self.n_users + 1, self.embed_size],
             initializer=tf_truncated_normal(0.0, 0.03),
             regularizer=tf_reg
         )
         self.item_embed_var = tf.get_variable(
             name="item_embed_var",
-            shape=[self.n_items, self.embed_size],
+            shape=[self.n_items + 1, self.embed_size],
             initializer=tf_truncated_normal(0.0, 0.03),
             regularizer=tf_reg
         )
@@ -150,20 +152,27 @@ class BPR(Base, TfMixin, EvalMixin):
         self.sess.run(tf.global_variables_initializer())
 
     def fit(self, train_data, verbose=1, shuffle=True, num_threads=1,
-            eval_data=None, metrics=None, optimizer="sgd"):
+            eval_data=None, metrics=None, optimizer="sgd", **kwargs):
         self.show_start_time()
         self._check_has_sampled(train_data, verbose)
 
+        if not self.graph_built and not self.model_built:
+            if self.use_tf:
+                self._build_model_tf()
+                self._build_train_ops()
+            else:
+                self._build_model()
+
         if self.use_tf:
             self._fit_tf(train_data, verbose=verbose, shuffle=shuffle,
-                         eval_data=eval_data, metrics=metrics)
+                         eval_data=eval_data, metrics=metrics, **kwargs)
         else:
             self._fit_cython(train_data, verbose=verbose, shuffle=shuffle,
                              num_threads=num_threads, eval_data=eval_data,
-                             metrics=metrics, optimizer=optimizer)
+                             metrics=metrics, optimizer=optimizer, **kwargs)
 
     def _fit_cython(self, train_data, verbose=1, shuffle=True, num_threads=1,
-                    eval_data=None, metrics=None, optimizer="sgd"):
+                    eval_data=None, metrics=None, optimizer="sgd", **kwargs):
         if optimizer == "sgd":
             trainer = partial(bpr_update)
 
@@ -212,11 +221,13 @@ class BPR(Base, TfMixin, EvalMixin):
                         epoch=epoch)
 
             if verbose > 1:
-                self.print_metrics(eval_data=eval_data, metrics=metrics)
+                self.print_metrics(eval_data=eval_data, metrics=metrics,
+                                   **kwargs)
                 print("="*30)
+        assign_oov_vector(self)
 
     def _fit_tf(self, train_data, verbose=1, shuffle=True,
-                eval_data=None, metrics=None):
+                eval_data=None, metrics=None, **kwargs):
         data_generator = PairwiseSampling(train_data,
                                           self.data_info,
                                           self.num_neg)
@@ -232,25 +243,18 @@ class BPR(Base, TfMixin, EvalMixin):
                                              self.item_indices_neg: item_neg})
 
             if verbose > 1:
-                # set up parameters for evaluate
+                # set up parameters for evaluation
                 self._set_latent_factors()
-                self.print_metrics(eval_data=eval_data, metrics=metrics)
+                self.print_metrics(eval_data=eval_data, metrics=metrics,
+                                   **kwargs)
                 print("="*30)
 
         # for prediction and recommendation
+        self.assign_oov()
         self._set_latent_factors()
 
-    def predict(self, user, item):
-        user = (
-            np.asarray([user])
-            if isinstance(user, int)
-            else np.asarray(user)
-        )
-        item = (
-            np.asarray([item])
-            if isinstance(item, int)
-            else np.asarray(item)
-        )
+    def predict(self, user, item, cold_start="average", inner_id=False):
+        user, item = self.convert_id(user, item, inner_id)
         unknown_num, unknown_index, user, item = self._check_unknown(user, item)
 
         preds = np.sum(
@@ -259,28 +263,33 @@ class BPR(Base, TfMixin, EvalMixin):
         )
         preds = 1 / (1 + np.exp(-preds))
 
-        if unknown_num > 0:
+        if unknown_num > 0 and cold_start == "popular":
             preds[unknown_index] = self.default_prediction
+        return preds
 
-        return preds[0] if len(user) == 1 else preds
+    def recommend_user(self, user, n_rec, cold_start="average", inner_id=False):
+        user_id = self._check_unknown_user(user, inner_id)
+        if user_id is None:
+            if cold_start == "average":
+                user_id = self.n_users
+            elif cold_start == "popular":
+                return self.data_info.popular_items[:n_rec]
+            else:
+                raise ValueError(user)
 
-    def recommend_user(self, user, n_rec, **kwargs):
-        user = self._check_unknown_user(user)
-        if not user:
-            return   # popular ?
-
-        consumed = self.user_consumed[user]
+        consumed = set(self.user_consumed[user_id])
         count = n_rec + len(consumed)
-        recos = self.user_embed[user] @ self.item_embed.T
+        recos = self.user_embed[user_id] @ self.item_embed.T
         recos = 1 / (1 + np.exp(-recos))
 
         ids = np.argpartition(recos, -count)[-count:]
         rank = sorted(zip(ids, recos[ids]), key=lambda x: -x[1])
-        return list(
-            islice(
-                (rec for rec in rank if rec[0] not in consumed), n_rec
-            )
+        recs_and_scores = islice(
+            (rec if inner_id else (self.data_info.id2item[rec[0]], rec[1])
+             for rec in rank if rec[0] not in consumed),
+            n_rec
         )
+        return list(recs_and_scores)
 
     def _set_latent_factors(self):
         item_bias, user_embed, item_embed = self.sess.run(
@@ -293,3 +302,26 @@ class BPR(Base, TfMixin, EvalMixin):
         item_bias = item_bias[:, None]
         self.user_embed = np.hstack([user_embed, user_bias])
         self.item_embed = np.hstack([item_embed, item_bias])
+
+    def save(self, path, model_name, manual=True, inference_only=False):
+        if not os.path.isdir(path):
+            print(f"file folder {path} doesn't exists, creating a new one...")
+            os.makedirs(path)
+        self.save_params(path)
+        if inference_only:
+            variable_path = os.path.join(path, model_name)
+            np.savez_compressed(variable_path,
+                                user_embed=self.user_embed,
+                                item_embed=self.item_embed)
+        else:
+            self.save_variables(path, model_name, inference_only=False)
+
+    @classmethod
+    def load(cls, path, model_name, data_info, manual=True):
+        variable_path = os.path.join(path, f"{model_name}.npz")
+        variables = np.load(variable_path)
+        hparams = cls.load_params(path, data_info)
+        model = cls(**hparams)
+        model.user_embed = variables["user_embed"]
+        model.item_embed = variables["item_embed"]
+        return model
