@@ -6,14 +6,16 @@ Reference: Guorui Zhou et al.  "Deep Interest Network for Click-Through Rate Pre
 author: massquantity
 
 """
+import os
 from itertools import islice
 import numpy as np
-import tensorflow as tf2
+import pandas as pd
+import tensorflow.compat.v1 as tf
 from tensorflow.keras.initializers import (
     truncated_normal as tf_truncated_normal
 )
 from .base import Base, TfMixin
-from ..evaluate.evaluate import EvalMixin
+from ..evaluation.evaluate import EvalMixin
 from ..utils.tf_ops import (
     reg_config,
     dropout_config,
@@ -26,13 +28,19 @@ from ..utils.misc import time_block, colorize
 from ..utils.misc import count_params
 from ..feature import (
     get_predict_indices_and_values,
-    get_recommend_indices_and_values
+    get_recommend_indices_and_values,
+    features_from_dict,
+    add_item_features
 )
-tf = tf2.compat.v1
 tf.disable_v2_behavior()
 
 
 class DIN(Base, TfMixin, EvalMixin):
+    user_variables = ["user_feat"]
+    item_variables = ["item_feat"]
+    sparse_variables = ["sparse_feat"]
+    dense_variables = ["dense_feat"]
+
     def __init__(
             self,
             task,
@@ -56,7 +64,7 @@ class DIN(Base, TfMixin, EvalMixin):
     ):
         Base.__init__(self, task, data_info, lower_upper_bound)
         TfMixin.__init__(self, tf_sess_config)
-        EvalMixin.__init__(self, task)
+        EvalMixin.__init__(self, task, data_info)
 
         self.task = task
         self.data_info = data_info
@@ -104,8 +112,10 @@ class DIN(Base, TfMixin, EvalMixin):
             self.item_dense_col_indices = data_info.item_dense_col.index
         self.user_last_interacted = None
         self.last_interacted_len = None
+        self.all_args = locals()
 
     def _build_model(self):
+        self.graph_built = True
         tf.set_random_seed(self.seed)
         self.concat_embed, self.item_embed, self.seq_embed = [], [], []
         self._build_placeholders()
@@ -147,12 +157,12 @@ class DIN(Base, TfMixin, EvalMixin):
     def _build_variables(self):
         self.user_feat = tf.get_variable(
             name="user_feat",
-            shape=[self.n_users, self.embed_size],
+            shape=[self.n_users + 1, self.embed_size],
             initializer=tf_truncated_normal(0.0, 0.01),
             regularizer=self.reg)
         self.item_feat = tf.get_variable(
             name="item_feat",
-            shape=[self.n_items, self.embed_size],
+            shape=[self.n_items + 1, self.embed_size],
             initializer=tf_truncated_normal(0.0, 0.01),
             regularizer=self.reg)
         if self.sparse:
@@ -312,7 +322,7 @@ class DIN(Base, TfMixin, EvalMixin):
             pooled_outputs = attention_scores @ keys
             return pooled_outputs
 
-    def _build_train_ops(self, global_steps=None):
+    def _build_train_ops(self, **kwargs):
         if self.task == "rating":
             self.loss = tf.losses.mean_squared_error(labels=self.labels,
                                                      predictions=self.output)
@@ -328,6 +338,13 @@ class DIN(Base, TfMixin, EvalMixin):
         else:
             total_loss = self.loss
 
+        if self.lr_decay:
+            n_batches = int(self.data_info.data_size / self.batch_size)
+            self.lr, global_steps = lr_decay_config(self.lr, n_batches,
+                                                    **kwargs)
+        else:
+            global_steps = None
+
         optimizer = tf.train.AdamOptimizer(self.lr)
         optimizer_op = optimizer.minimize(total_loss, global_step=global_steps)
         update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
@@ -337,15 +354,9 @@ class DIN(Base, TfMixin, EvalMixin):
     def fit(self, train_data, verbose=1, shuffle=True,
             eval_data=None, metrics=None, **kwargs):
         self.show_start_time()
-        if self.lr_decay:
-            n_batches = int(len(train_data) / self.batch_size)
-            self.lr, global_steps = lr_decay_config(self.lr, n_batches,
-                                                    **kwargs)
-        else:
-            global_steps = None
-
-        self._build_model()
-        self._build_train_ops(global_steps)
+        if not self.graph_built:
+            self._build_model()
+            self._build_train_ops(**kwargs)
 
         data_generator = DataGenSequence(train_data, self.data_info,
                                          self.sparse, self.dense,
@@ -375,31 +386,34 @@ class DIN(Base, TfMixin, EvalMixin):
                 print(f"\t {colorize(train_loss_str, 'green')}")
                 # for evaluation
                 self._set_last_interacted()
-                self.print_metrics(eval_data=eval_data, metrics=metrics)
+                self.print_metrics(eval_data=eval_data, metrics=metrics,
+                                   **kwargs)
                 print("=" * 30)
 
         # for prediction and recommendation
         self._set_last_interacted()
+        self.assign_oov()
 
-    def predict(self, user, item):
-        user = (
-            np.asarray([user])
-            if isinstance(user, int)
-            else np.asarray(user)
-        )
-        item = (
-            np.asarray([item])
-            if isinstance(item, int)
-            else np.asarray(item)
-        )
+    def predict(self, user, item, feats=None, cold_start="average",
+                inner_id=False):
+        user, item = self.convert_id(user, item, inner_id)
         unknown_num, unknown_index, user, item = self._check_unknown(user, item)
 
-        (user_indices,
-         item_indices,
-         sparse_indices,
-         dense_values) = get_predict_indices_and_values(
-            self.data_info, user, item, self.n_items, self.sparse, self.dense
-        )
+        (
+            user_indices,
+            item_indices,
+            sparse_indices,
+            dense_values
+        ) = get_predict_indices_and_values(
+            self.data_info, user, item, self.n_items, self.sparse, self.dense)
+
+        if feats is not None:
+            assert isinstance(feats, (dict, pd.Series)), (
+                "feats must be dict or pandas.Series.")
+            assert len(user_indices) == 1, "only support single user for feats"
+            sparse_indices, dense_values = features_from_dict(
+                self.data_info, sparse_indices, dense_values, feats, "predict")
+
         feed_dict = self._get_seq_feed_dict(self.user_last_interacted[user],
                                             self.last_interacted_len[user],
                                             user_indices, item_indices,
@@ -407,47 +421,99 @@ class DIN(Base, TfMixin, EvalMixin):
                                             dense_values, False)
 
         preds = self.sess.run(self.output, feed_dict)
-        preds = 1 / (1 + np.exp(-preds))
-        if unknown_num > 0:
+        if self.task == "rating":
+            preds = np.clip(preds, self.lower_bound, self.upper_bound)
+        elif self.task == "ranking":
+            preds = 1 / (1 + np.exp(-preds))
+
+        if unknown_num > 0 and cold_start == "popular":
             preds[unknown_index] = self.default_prediction
+        return preds
 
-        return preds[0] if len(user) == 1 else preds
+    def recommend_user(self, user, n_rec, user_feats=None, item_data=None,
+                       cold_start="average", inner_id=False):
+        user_id = self._check_unknown_user(user, inner_id)
+        if user_id is None:
+            if cold_start == "average":
+                user_id = self.n_users
+            elif cold_start == "popular":
+                return self.data_info.popular_items[:n_rec]
+            else:
+                raise ValueError(user)
 
-    def recommend_user(self, user, n_rec, **kwargs):
-        user = self._check_unknown_user(user)
-        if not user:
-            return   # popular ?
+        (
+            user_indices,
+            item_indices,
+            sparse_indices,
+            dense_values
+        ) = get_recommend_indices_and_values(
+            self.data_info, user_id, self.n_items, self.sparse, self.dense)
 
-        (user_indices,
-         item_indices,
-         sparse_indices,
-         dense_values) = get_recommend_indices_and_values(
-            self.data_info, user, self.n_items, self.sparse, self.dense)
-        u_last_interacted = np.tile(self.user_last_interacted[user],
+        if user_feats is not None:
+            assert isinstance(user_feats, (dict, pd.Series)), (
+                "feats must be dict or pandas.Series.")
+            sparse_indices, dense_values = features_from_dict(
+                self.data_info, sparse_indices, dense_values, user_feats,
+                "recommend")
+        if item_data is not None:
+            assert isinstance(item_data, pd.DataFrame), (
+                "item_data must be pandas DataFrame")
+            assert "item" in item_data.columns, (
+                "item_data must contain 'item' column")
+            sparse_indices, dense_values = add_item_features(
+                self.data_info, sparse_indices, dense_values, item_data)
+
+        u_last_interacted = np.tile(self.user_last_interacted[user_id],
                                     (self.n_items, 1))
-        u_interacted_len = np.repeat(self.last_interacted_len[user],
+        u_interacted_len = np.repeat(self.last_interacted_len[user_id],
                                      self.n_items)
         feed_dict = self._get_seq_feed_dict(u_last_interacted, u_interacted_len,
                                             user_indices, item_indices, None,
                                             sparse_indices, dense_values, False)
 
         recos = self.sess.run(self.output, feed_dict)
-        recos = 1 / (1 + np.exp(-recos))
-        consumed = self.user_consumed[user]
+        if self.task == "ranking":
+            recos = 1 / (1 + np.exp(-recos))
+        consumed = set(self.user_consumed[user_id])
         count = n_rec + len(consumed)
         ids = np.argpartition(recos, -count)[-count:]
         rank = sorted(zip(ids, recos[ids]), key=lambda x: -x[1])
-        return list(
-            islice(
-                (rec for rec in rank if rec[0] not in consumed), n_rec
-            )
+        recs_and_scores = islice(
+            (rec if inner_id else (self.data_info.id2item[rec[0]], rec[1])
+             for rec in rank if rec[0] not in consumed),
+            n_rec
         )
+        return list(recs_and_scores)
 
     def _set_last_interacted(self):
         if (self.user_last_interacted is None
                 and self.last_interacted_len is None):
             user_indices = np.arange(self.n_users)
-            (self.user_last_interacted,
-             self.last_interacted_len) = user_last_interacted(
-                user_indices, self.user_consumed, 0, self.max_seq_len)
+            (
+                self.user_last_interacted,
+                self.last_interacted_len
+            ) = user_last_interacted(user_indices, self.user_consumed,
+                                     self.n_items, self.max_seq_len)
 
+            oov = np.full(self.max_seq_len, self.n_items, dtype=np.int32)
+            self.user_last_interacted = np.vstack(
+                [self.user_last_interacted, oov]
+            )
+            self.last_interacted_len = np.append(self.last_interacted_len, [1])
+
+    def save(self, path, model_name, manual=True, inference_only=False):
+        if not os.path.isdir(path):
+            print(f"file folder {path} doesn't exists, creating a new one...")
+            os.makedirs(path)
+        self.save_params(path)
+        if manual:
+            self.save_variables(path, model_name, inference_only)
+        else:
+            self.save_tf_model(path, model_name)
+
+    @classmethod
+    def load(cls, path, model_name, data_info, manual=True):
+        if manual:
+            return cls.load_variables(path, model_name, data_info)
+        else:
+            return cls.load_tf_model(path, model_name, data_info)
